####################################################################
# Script in order to run multivariate association analysis using   #
# CHARGE+UKBB single macronutrient intake GWAS results             #
# Chloe Sarnowski, Chloe.Sarnowski@uth.tmc.edu                     #
# 08/02/21                                                         #
####################################################################

library('data.table')
library("MASS")
library("Matrix")
require(compiler)
enableJIT(4)

# Open each macronutrient intake GWAS results and merge them
d1=fread("carb_ukbb_envt.txt",header=T)
d2=fread("fat_ukbb_envt.txt",header=T)
d3=fread("prot_ukbb_envt.txt",header=T)

d1_snps <- subset(d1, select=c('SNPID'))
d2_snps <- subset(d2, select=c('SNPID'))
d3_snps <- subset(d3, select=c('SNPID'))

all_snps <- merge(d1_snps, d2_snps, by='SNPID')
all_snps <- merge(all_snps, d3_snps, by='SNPID')

d1 <- merge(d1, all_snps, by='SNPID')
d2 <- merge(d2, all_snps, by='SNPID')
d3 <- merge(d3, all_snps, by='SNPID')

d1 <- subset(d1, select=c('SNPID', 'Allele1', 'BETA', 'SE', 'p.value'))
d2 <- subset(d2, select=c('SNPID', 'Allele1', 'BETA', 'SE', 'p.value'))
d3 <- subset(d3, select=c('SNPID', 'Allele1', 'BETA', 'SE', 'p.value'))

names(d1) <- c("SNP","ALLELE1","BETA","SE","P")
names(d2) <- c("SNP","ALLELE1","BETA","SE","P")
names(d3) <- c("SNP","ALLELE1","BETA","SE","P")

#an example of R code showing how to estimate the correlation matrix.
# read the data: the summary statistics data file (“HeightBMIWHRadjBMI.txt”) for CPASSOC includes seven columns (SNP ID, height Z-score for male, height Z-score for female, BMI Z-score for male, BMI Z-score for female, WHRadjBMI Z-score for male, WHRadjBMI Z-score for female. The SNP list file after pruning based on ARIC data is “ARICdot2.prune.in”.
#Dt<-read.table(“HeightBMIWHRadjBMI.txt”,header=T)
# extract the SNP set obtained from ARIC EA data.
#pruned.lst<-read.table(“ARICdot2.prune.in”)
#Dtcorr<-Dt[Dt$MarkerName%in%pruned.lst$V1,]
# Remove SNPs with Z-score larger than 1.96 or smaller than −1.96, assuming the summary statistics for 6 sex-specific traits are named Z1,Z2,…,Z6.
#Index<-which(abs(Dtcorr$Z1)>1.96 | abs(Dtcorr$Z2)>1.96 |abs(Dtcorr$Z3)>1.96 | abs(Dtcorr$Z4)>1.96 |abs(Dtcorr$Z5)> 1.96 | abs(Dtcorr$Z6)> 1.96)
#Dtcorr<-Dtcorr[-Index,]
#corMatrix<-diag(x=1,ncol=6,nrow=6)
#for(i in 1:5){
#for(j in (i+1):6)
#corMatrix[i,j]=corMatrix[j,i]=cor(Dtcorr[,i+1],Dtcorr[,j+1])
#}

## Please make sure d1,d2,d3 have the same order of SNPs and tested allele ##

data=data.frame(d1$BETA/d1$SE,d2$BETA/d2$SE,d3$BETA/d3$SE)
# Sample size to be defined by the user
S=rep(191157,times=3)
# Correlation matrix provided by the user (here estimated in the UKBB)
B=matrix(c(1,-0.36,-0.14,-0.36,1,0.007,-0.14,0.007,1),3,3) 

Non_Trucated_TestScore <- function(X, SampleSize, CorrMatrix) 
{
  Wi = matrix(SampleSize, nrow = 1);
  sumW = sqrt(sum(Wi^2));
  W = Wi / sumW;
  
  Sigma = ginv(CorrMatrix);
  XX = apply(X, 1, function(x) {
    x1 <- matrix(x, ncol = length(x), nrow = 1); 
    T = W %*% Sigma %*% t(x1);
    T = (T*T) / (W %*% Sigma %*% t(W));
    return(T[1,1]);
  }		
  );		
  return(XX);
}
SHom <- cmpfun(Non_Trucated_TestScore);

Trucated_TestScore <- function(X, SampleSize, CorrMatrix, correct = 1, startCutoff = 0, endCutoff = 1, CutoffStep = 0.05, isAllpossible = T) 
{
  N = dim(X)[2];
  
  Wi = matrix(SampleSize, nrow = 1);
  sumW = sqrt(sum(Wi^2));
  W = Wi / sumW;	
  
  XX = apply(X, 1, function(x) {
    TTT = -1;
    
    if (isAllpossible == T ) {
      cutoff = sort(unique(abs(x)));	  ## it will filter out any of them. 
    } else {
      cutoff = seq(startCutoff, endCutoff, CutoffStep);		
    }
    
    
    for (threshold in cutoff) {
      x1 = x;
      index = which(abs(x1) < threshold);
      
      if (length(index) == N) break;
      
      A = CorrMatrix;
      
      W1 = W;	
      if (length(index) !=0 ) {   
        x1 = x1[-index];						
        A  = A[-index, -index];   ## update the matrix
        W1 = W[-index]; 
      }
      
      if (correct == 1)
      {
        index = which(x1 < 0);
        if (length(index) != 0) {
          W1[index] = -W1[index];    ## update the sign
        }
      }
      
      A = ginv(A);
      x1 = matrix(x1, nrow = 1); 
      W1 = matrix(W1, nrow = 1);
      T = W1 %*% A %*% t(x1);
      T = (T*T) / (W1 %*% A %*% t(W1));
      
      if (TTT < T[1,1]) TTT = T[1,1];	
    }
    return(TTT);
  }		
  );		
  return(XX);
}
SHet <- cmpfun(Trucated_TestScore);

EstimateGamma <- function (N = 1E6, SampleSize, CorrMatrix, correct = 1, startCutoff = 0, endCutoff = 1, CutoffStep = 0.05, isAllpossible = T) {
  
  Wi = matrix(SampleSize, nrow = 1);   
  sumW = sqrt(sum(Wi^2));
  W = Wi / sumW;
  
  Permutation = mvrnorm(n = N, mu = c(rep(0, length(SampleSize))), Sigma = CorrMatrix, tol = 1e-8, empirical = F);
  
  Stat =  Trucated_TestScore(X = Permutation, SampleSize = SampleSize, CorrMatrix = CorrMatrix, 
                             correct = correct, startCutoff = startCutoff, endCutoff = endCutoff, 
                             CutoffStep = CutoffStep, isAllpossible = isAllpossible);
  a = min(Stat)*3/4
  ex3 = mean(Stat*Stat*Stat)
  V =	var(Stat);
  
  for (i in 1:100){
    E = mean(Stat)-a;
    k = E^2/V
    theta = V/E
    a = (-3*k*(k+1)*theta**2+sqrt(9*k**2*(k+1)**2*theta**4-12*k*theta*(k*(k+1)*(k+2)*theta**3-ex3)))/6/k/theta
  }
  
  para = c(k,theta,a);
  return(para);
}	  

EmpDist <- function (N = 1E6, SampleSize, CorrMatrix, correct = 1, startCutoff = 0, endCutoff = 1, CutoffStep = 0.05, isAllpossible = T) {
  
  Wi = matrix(SampleSize, nrow = 1);   
  sumW = sqrt(sum(Wi^2));
  W = Wi / sumW;
  
  Permutation = mvrnorm(n = N, mu = c(rep(0, length(SampleSize))), Sigma = CorrMatrix, tol = 1e-8, empirical = F);
  
  Stat =  Trucated_TestScore(X = Permutation, SampleSize = SampleSize, CorrMatrix = CorrMatrix, correct = correct, startCutoff = startCutoff, endCutoff = endCutoff, CutoffStep = CutoffStep, isAllpossible = isAllpossible);
    
  return(Stat);
}	


para=EstimateGamma(N=10000,SampleSize=S,CorrMatrix=B)


#####################################################

shom.chisq=SHom(X=data,SampleSize=S,CorrMatrix=B)
shom.p=pchisq(shom.chisq,df=1,ncp=0,lower.tail=F)

shet.chisq=SHet(X=data,SampleSize=S,CorrMatrix=B)
shet.p=pgamma(q=shet.chisq-para[3],shape=para[1],scale=para[2],lower.tail=F)

y=data.frame(SNP=d1$SNP,A1=d1$ALLELE1,carb.p=d1$P,fat.p=d2$P,pro.p=d3$P,SHom.chisq=shom.chisq,SHom.p=shom.p,SHet.chisq=shet.chisq,SHet.p=shet.p)

write.table(y,"UKB_Macronutrient_CPASSOC_Env2.txt", quote=F,col.names=T,row.names=F,sep="\t")
